/*
 * UAE - The Ultimate Amiga Emulator
 *
 * Win32 DirectX Wrappers, to simplify (?) my life.
 *
 * Copyright 1999 Brian King, under GNU Public License
 *
 */
#define INITGUID

#include "config.h"
#include "sysconfig.h"

#include <stdlib.h>
#include <stdarg.h>
#include <signal.h>
#include <windows.h>
#include <mmsystem.h>

#ifdef _MSC_VER
#include <ddraw.h>
#include <dsound.h>
#else
#include "winstuff.h"
#endif

#include "sysdeps.h"
#include "options.h"
#include "picasso96.h"
#include "osdep/dxwrap.h"
#include "osdep/win32.h"

extern HRESULT (WINAPI * pDirectDrawCreate) (GUID FAR *, LPDIRECTDRAW FAR *, IUnknown FAR *);

/* Uncomment the following line for simulation of non-lockable front-buffer (primary-surface) */
/* #define TEST_SECONDARY_SURFACE */

#ifdef __GNUC__
/* We'll only QueryInterface for the highest versions */
DEFINE_GUID( IID_IDirectDraw4,        0x9c59509a,0x39bd,0x11d1,0x8c,0x4a,0x00,0xc0,0x4f,0xd9,0x30,0xc5 );
DEFINE_GUID( IID_IDirectDrawSurface4, 0x0B2B8630,0xAD35,0x11D0,0x8E,0xA6,0x00,0x60,0x97,0x97,0xEA,0x5B );
#endif

/*
 * FUNCTION:   ShowDDCaps
 *
 * PURPOSE:    print out the DirectDraw Capabilities
 *
 * PARAMETERS:
 *   caps    - DDCAPS structure
 *   hw      - flag indicating if this 'caps' is for real hardware or the HEL
 *
 * RETURNS:    none
 *
 * NOTES:      none
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
static void ShowDDCaps( DDCAPS caps, int hw )
{
    static int shown = 0;

    if( shown < 2)
    {
	    write_log( "DirectDraw Capabilities for %s:\n", hw ? "Display Driver Hardware" : "Display Driver Emulation Layer" );
	    if( caps.dwCaps & DDCAPS_BLT )
		    write_log( "DDCAPS_BLT - Capable of blitting\n" );
	    if( caps.dwCaps & DDCAPS_BLTCOLORFILL )
		    write_log( "DDCAPS_BLTCOLORFILL - Color filling with blitter\n" );
	    if( caps.dwCaps & DDCAPS_BLTSTRETCH )
		    write_log( "DDCAPS_BLTSTRETCH - Stretch blitting\n" );
	    if( caps.dwCaps & DDCAPS_CANBLTSYSMEM )
		    write_log( "DDCAPS_CANBLTSYSMEM - Blits from system memory\n" );
	    if( caps.dwCaps & DDCAPS_CANCLIP )
		    write_log( "DDCAPS_CANCLIP - Can clip while blitting\n" );
	    if( caps.dwCaps & DDCAPS_CANCLIPSTRETCHED )
		    write_log( "DDCAPS_CANCLIPSTRETCHED - Can clip while stretch-blitting\n" );
	    if( caps.dwCaps & DDCAPS_GDI )
		    write_log( "DDCAPS_GDI - Display h/w shared with GDI\n" );
	    if( caps.dwCaps & DDCAPS_NOHARDWARE )
		    write_log( "DDCAPS_NOHARDWARE - no h/w support!\n" );
	    
	    if( caps.dwCaps2 & DDCAPS2_CERTIFIED )
		    write_log( "DDCAPS2_CERTIFIED - certified driver\n" );
            if( caps.dwCaps2 & DDCAPS2_CANRENDERWINDOWED )
                write_log( "DDCAPS2_CANRENDERWINDOWED - GDI windows can be seen when in full-screen\n" );
	    if( caps.dwCaps2 & DDCAPS2_NOPAGELOCKREQUIRED )
		    write_log( "DDCAPS2_NOPAGELOCKREQUIRED - no page locking needed for DMA blits\n" );

	    write_log( "Video memory: %d/%d\n", caps.dwVidMemFree, caps.dwVidMemTotal );

        shown++;
    }
}

/*
 * FUNCTION:   DirectDraw_ErrorString
 *
 * PURPOSE:    Convert a DirectX error value to a text-string
 *
 * PARAMETERS:
 *   ddrval  - HRESULT of a failed call
 *
 * RETURNS:    ASCII string
 *
 * NOTES:      none
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
char *DirectDraw_ErrorString( HRESULT ddrval )
{
    switch (ddrval) 
    {
    case DDERR_ALREADYINITIALIZED:
	    return "This object is already initialized.";
    case DDERR_CANNOTATTACHSURFACE:
	    return "Cannot attach surface.";
    case DDERR_CANNOTDETACHSURFACE:
	    return "Cannot detach surface.";
    case DDERR_CURRENTLYNOTAVAIL:
	    return "Support unavailable.";
    case DDERR_EXCEPTION:
	    return "Unexpected exception.";
    case DDERR_GENERIC:
	    return "Undefined";
    case DDERR_HEIGHTALIGN:
	    return "Height needs to be aligned.";
    case DDERR_INCOMPATIBLEPRIMARY:
	    return "New params doesn't match existing primary surface.";
    case DDERR_INVALIDCAPS:
	    return "Device doesn't have requested capabilities.";
    case DDERR_INVALIDCLIPLIST:
	    return "Provided clip-list not supported.";
    case DDERR_INVALIDMODE:
	    return "Mode not supported.";
    case DDERR_INVALIDOBJECT:
	    return "Invalid object.";
    case DDERR_INVALIDPARAMS:
	    return "Invalid params.";
    case DDERR_INVALIDPIXELFORMAT:
	    return "Device doesn't support requested pixel format.";
    case DDERR_INVALIDRECT:
	    return "Invalid RECT.";
    case DDERR_LOCKEDSURFACES:
	    return "Surface locked.";
    case DDERR_NO3D:
	    return "No 3d capabilities.";
    case DDERR_NOALPHAHW:
	    return "No alpha h/w.";
    case DDERR_NOCLIPLIST:
	    return "No clip-list.";
    case DDERR_NOCOLORCONVHW:
	    return "No colour-conversion h/w.";
    case DDERR_NOCOOPERATIVELEVELSET:
        return "No cooperative-level set.";
    case DDERR_NOCOLORKEY:
	    return "No colour-key.";
    case DDERR_NOCOLORKEYHW:
        return "No colour-key hardware.";
    case DDERR_NODIRECTDRAWSUPPORT:
        return "No DirectDraw support with this display driver!";
    case DDERR_NOEXCLUSIVEMODE:
        return "Exlusive-mode needed but not set yet.";
    case DDERR_NOFLIPHW:
        return "No flipping hardware.";
    case DDERR_NOGDI:
        return "No GDI present.";
    case DDERR_NOMIRRORHW:
        return "No mirror hardware.";
    case DDERR_NOTFOUND:
        return "Requested item not found.";
    case DDERR_NOOVERLAYHW:
        return "No overlay hardware.";
    case DDERR_NORASTEROPHW:
        return "No raster-op hardware.";
    case DDERR_NOROTATIONHW:
        return "No rotation hardware.";
    case DDERR_NOSTRETCHHW:
        return "No stretch hardware.";
    case DDERR_NOT4BITCOLOR:
        return "Not a 4-bit colour palette.";
    case DDERR_NOT4BITCOLORINDEX:
        return "Not a 4-bit colour-index.";
    case DDERR_NOT8BITCOLOR:
        return "Not an 8-bit colour palette.";
    case DDERR_NOTEXTUREHW:
        return "No texture hardware.";
    case DDERR_NOVSYNCHW:
        return "No vsync hardware.";
    case DDERR_NOZBUFFERHW:
        return "No z-buffer hardware.";
    case DDERR_NOZOVERLAYHW:
        return "No z-overlay hardware.";
    case DDERR_OUTOFCAPS:
    	return "out of caps";
    case DDERR_OUTOFMEMORY:
	    return "Out of memory.";
    case DDERR_OUTOFVIDEOMEMORY:
	    return "out of video memory.";
    case DDERR_OVERLAYCANTCLIP:
        return "no clipped overlay support in h/w.";
    case DDERR_OVERLAYCOLORKEYONLYONEACTIVE:
        return "can only have one color-key active at one time for overlays.";
    case DDERR_PALETTEBUSY:
    	return "Palette busy.";
    case DDERR_COLORKEYNOTSET:
    	return "No source colour-key provided.";
    case DDERR_SURFACEALREADYATTACHED:
        return "Surface already attached.";
    case DDERR_SURFACEALREADYDEPENDENT:
        return"Surface already dependent.";
    case DDERR_SURFACEBUSY:
    	return "Surface busy.";
    case DDERR_CANTLOCKSURFACE:
    	return "Access to surface refused (no DCI Provider).";
    case DDERR_SURFACEISOBSCURED:
        return "Surface is obscured.";
    case DDERR_SURFACELOST:
    	return "Surface lost.";
    case DDERR_SURFACENOTATTACHED:
        return "Surface not attached.";
    case DDERR_TOOBIGHEIGHT:
        return "Height requested is too large.";
    case DDERR_TOOBIGSIZE:
        return "Size requested is too large.";
    case DDERR_TOOBIGWIDTH:
        return "Width requested is too large.";
    case DDERR_UNSUPPORTED:
        return "Unsupported.";
    case DDERR_UNSUPPORTEDFORMAT:
    	return "Unsupported format.";
    case DDERR_UNSUPPORTEDMASK:
        return "Unsupported bitmask in pixel-format request.";
    case DDERR_VERTICALBLANKINPROGRESS:
        return "Vertical blank in progress.";
    case DDERR_WASSTILLDRAWING:
    	return "Was still drawing.";
    case DDERR_XALIGN:
        return "rectangle needs to be x-aligned.";
    case DDERR_INVALIDDIRECTDRAWGUID:
        return "invalid GUID for direct-draw.";
    case DDERR_DIRECTDRAWALREADYCREATED:
        return "DirectDraw object already created for this process.";
    case DDERR_NODIRECTDRAWHW:
        return "driver does not support h/w DirectDraw.";
    case DDERR_PRIMARYSURFACEALREADYEXISTS:
        return "Primary surface already exists for this process.";
    case DDERR_NOEMULATION:
        return "DirectDraw s/w emulation not available.";
    case DDERR_REGIONTOOSMALL:
        return "GetClipList() region is too small.";
    case DDERR_CLIPPERISUSINGHWND:
        return "Trying to set a clip-list for a clipper object which is already attached to an HWND.";
    case DDERR_NOCLIPPERATTACHED:
        return "No clipper object attached to surface object.";
    case DDERR_NOHWND:
        return "No HWND provided for clipper notification, or for CooperativeLevel.";
    case DDERR_HWNDSUBCLASSED:
        return "CooperativeLevel HWND has been subclassed, preventing state restoration.";
    case DDERR_HWNDALREADYSET:
        return "CooperativeLevel HWND already set - can't be reset while surfaces or palettes exist.";
    case DDERR_NOPALETTEATTACHED:
        return "No palette object attached to this surface.";
    case DDERR_NOPALETTEHW:
        return "No h/w support for 16 or 256 colour palettes.";
    case DDERR_BLTFASTCANTCLIP:
    	return "Cannot use BLTFAST with Clipper attached to surface.";
    case DDERR_NOBLTHW:
        return "No blitter.";
    case DDERR_NODDROPSHW:
        return "No DirectDraw ROP h/w.";
    case DDERR_OVERLAYNOTVISIBLE:
        return "GetOverlayPosition called on a hidden overlay.";
    case DDERR_NOOVERLAYDEST:
        return "GetOverlayPosition called on an overlay that UpdateOverlay has never been called on.";
    case DDERR_INVALIDPOSITION:
        return "Position of overlay on destination is no longer legal.";
    case DDERR_NOTAOVERLAYSURFACE:
        return "Overlay member is called for a non-overlay surface.";
    case DDERR_EXCLUSIVEMODEALREADYSET:
        return "Attempt to set CooperativeLevel when already set to exclusive.";
    case DDERR_NOTFLIPPABLE:
        return "Surface not flippable.";
    case DDERR_CANTDUPLICATE:
        return "Can't duplicate primary or 3d surfaces, or surfaces that are implicitly created.";
    case DDERR_NOTLOCKED:
        return "Surface not locked.";
    case DDERR_CANTCREATEDC:
        return "Cannot create any more DCs.";
    case DDERR_NODC:
        return "No DC ever created for this surface.";
    case DDERR_WRONGMODE:
        return "Surface cannot be restored, because it was created in a different mode.";
    case DDERR_IMPLICITLYCREATED:
        return "Surface cannot be restored, because it is an implicitly created surface.";
    case DDERR_NOTPALETTIZED:
        return "Surface being used is not a palette-based surface.";
    case DDERR_UNSUPPORTEDMODE:
        return "The display is currently in an unsupported mode.";
    case DDERR_NOMIPMAPHW:
        return "Operation could not be carried out because there is no mip-map texture mapping hardware present or available.";
    case DDERR_INVALIDSURFACETYPE:
        return "The requested action could not be performed because the surface was of the wrong type.";
    case DDERR_NOOPTIMIZEHW:
        return "Device does not support optimized surfaces, therefore no video memory optimized surfaces.";
    case DDERR_NOTLOADED:
        return "Surface is an optimized surface, but has not yet been allocated any memory.";
    case DDERR_NOFOCUSWINDOW:
        return "Attempt was made to create or set a device window without first setting the focus window.";
    case DDERR_DCALREADYCREATED:
        return "A DC has already been returned for this surface. Only one DC can be retrieved per surface.";
    case DDERR_NONONLOCALVIDMEM:
        return "An attempt was made to allocate non-local video memory from a device that does not support non-local video memory.";
    case DDERR_CANTPAGELOCK:
    	return "Attempt to page-lock a surface failed.";
    case DDERR_CANTPAGEUNLOCK:
	    return "Attempt to page-unlock a surface failed.";
    case DDERR_NOTPAGELOCKED:
        return "An attempt was made to page unlock a surface with no outstanding page locks.";
    case DDERR_MOREDATA:
        return "There is more data available than the specified buffer size could hold.";
    case DDERR_VIDEONOTACTIVE:
        return "The video port is not active.";
    case DDERR_DEVICEDOESNTOWNSURFACE:
        return "Surfaces created by one direct draw device cannot be used directly by another direct draw device.";
    case DDERR_NOTINITIALIZED:
        return "An attempt was made to invoke an interface member of a DirectDraw object created by CoCreateInstance() before it was initialized.";
    }
    return "";
}

static struct DirectDrawSurfaceMapper DirectDrawState;

static int lockcnt = 0;

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
static int LockStub4( LPDIRECTDRAWSURFACE4 surface, LPDDSURFACEDESC2 surfacedesc )
{
    int result = 0;
    HRESULT ddrval;

    while ( (ddrval = IDirectDrawSurface4_Lock( surface, NULL, surfacedesc, DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT,
                                                NULL ) ) != DD_OK )
    {
        if (ddrval == DDERR_SURFACELOST) {
    	    ddrval = IDirectDrawSurface4_Restore( surface );
            if (ddrval != DD_OK)
            {
                result = 0;
                break;
            }
        }
        else if (ddrval != DDERR_SURFACEBUSY) 
        {
	    write_log ("lpDDS->Lock() failed - %s (%d)\n", DirectDraw_ErrorString (ddrval), (unsigned short) ddrval);
            result = 0;
            break;
        }
    }
    if( ddrval == DD_OK )
        result = 1;
    
    if( result )
        lockcnt++;

    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
static int LockStub( LPDIRECTDRAWSURFACE surface, LPDDSURFACEDESC surfacedesc )
{
    int result = 0;
    HRESULT ddrval;

    while ( (ddrval = IDirectDrawSurface_Lock( surface, NULL, surfacedesc, DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT,
                                                NULL ) ) != DD_OK )
    {
        if (ddrval == DDERR_SURFACELOST) {
    	    ddrval = IDirectDrawSurface_Restore( surface );
            if (ddrval != DD_OK)
            {
                result = 0;
                break;
            }
        }
        else if (ddrval != DDERR_SURFACEBUSY) 
        {
	    write_log ("lpDDS->Lock() failed - %s (%d)\n", DirectDraw_ErrorString (ddrval), (unsigned short) ddrval);
            result = 0;
            break;
        }
    }
    if( ddrval == DD_OK )
        result = 1;
    
    if( result )
        lockcnt++;

    return result;
}

/* For a given surface-type, update our DirectDrawState structure */
/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
int DirectDraw_SurfaceLock( surface_type_e surface_type )
{
    int result = 0;

    if( surface_type == lockable_surface )
        surface_type = DirectDraw_GetLockableType();

    switch( surface_type )
    {
        case primary_surface:
            if( DirectDrawState.directdraw6 )
            {
                DirectDrawState.primary.desc_dx6.dwSize = sizeof( DDSURFACEDESC2 );
                result = LockStub4( DirectDrawState.primary.surface_dx6, &DirectDrawState.primary.desc_dx6 );
            }
            else
            {
                DirectDrawState.primary.desc.dwSize = sizeof( DDSURFACEDESC );
                result = LockStub( DirectDrawState.primary.surface, &DirectDrawState.primary.desc );
            }
        break;
        case secondary_surface:
            if( DirectDrawState.directdraw6 )
            {
                DirectDrawState.secondary.desc_dx6.dwSize = sizeof( DDSURFACEDESC2 );
                result = LockStub4( DirectDrawState.secondary.surface_dx6, &DirectDrawState.secondary.desc_dx6 );
            }
            else
            {
                DirectDrawState.secondary.desc.dwSize = sizeof( DDSURFACEDESC );
                result = LockStub( DirectDrawState.secondary.surface, &DirectDrawState.secondary.desc );
            }
        break;

        case lockable_surface:
        case invalid_surface:
        default:

        break;
    }
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
char *DirectDraw_GetSurfacePointer( void )
{
    char *pixels = NULL;

    /* Make sure that somebody has done a lock before returning the lpSurface member */
    if( lockcnt )
    {
        if( DirectDrawState.directdraw6 )
        {
            pixels = DirectDrawState.lockable.lpdesc_dx6->lpSurface;
        }
        else
        {
            pixels = DirectDrawState.lockable.lpdesc->lpSurface;
        }
    }
    return pixels;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
LONG DirectDraw_GetSurfacePitch( void )
{
    LONG pitch = 0;

    if( DirectDrawState.directdraw6 )
    {
        pitch = DirectDrawState.lockable.lpdesc_dx6->lPitch;
    }
    else
    {
        pitch = DirectDrawState.lockable.lpdesc->lPitch;
    }
    return pitch;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
DWORD DirectDraw_GetPixelFormatFlags( void )
{
    DWORD flags = 0;
    if( DirectDrawState.directdraw6 )
    {
        flags = DirectDrawState.lockable.lpdesc_dx6->ddpfPixelFormat.dwFlags;
    }
    else
    {
        flags = DirectDrawState.lockable.lpdesc->ddpfPixelFormat.dwFlags;
    }
    return flags;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
DWORD DirectDraw_GetSurfaceFlags( void )
{
    DWORD flags = 0;
    if( DirectDrawState.directdraw6 )
    {
        flags = DirectDrawState.lockable.lpdesc_dx6->dwFlags;
    }
    else
    {
        flags = DirectDrawState.lockable.lpdesc->dwFlags;
    }
    return flags;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
DWORD DirectDraw_GetSurfaceBitCount( void )
{
    DWORD bits = 0;
    if( DirectDrawState.directdraw6 )
    {
        bits = DirectDrawState.lockable.lpdesc_dx6->ddpfPixelFormat.dwRGBBitCount;
    }
    else
    {
        bits = DirectDrawState.lockable.lpdesc->ddpfPixelFormat.dwRGBBitCount;
    }
    return bits;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
DWORD DirectDraw_GetPixelFormatBitMask( DirectDraw_Mask_e mask )
{
    DWORD result = 0;
    switch( mask )
    {
        case red_mask:
            if( DirectDrawState.directdraw6 )
            {
                result = DirectDrawState.lockable.lpdesc_dx6->ddpfPixelFormat.dwRBitMask;
            }
            else
            {
                result = DirectDrawState.lockable.lpdesc->ddpfPixelFormat.dwRBitMask;
            }
        break;
        case green_mask:
            if( DirectDrawState.directdraw6 )
            {
                result = DirectDrawState.lockable.lpdesc_dx6->ddpfPixelFormat.dwGBitMask;
            }
            else
            {
                result = DirectDrawState.lockable.lpdesc->ddpfPixelFormat.dwGBitMask;
            }
        break;
        case blue_mask:
            if( DirectDrawState.directdraw6 )
            {
                result = DirectDrawState.lockable.lpdesc_dx6->ddpfPixelFormat.dwBBitMask;
            }
            else
            {
                result = DirectDrawState.lockable.lpdesc->ddpfPixelFormat.dwBBitMask;
            }
        break;
    }
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
surface_type_e DirectDraw_GetLockableType( void )
{
    return DirectDrawState.surface_type;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
surface_type_e try_surface_locks( int want_fullscreen )
{
    surface_type_e result = invalid_surface;
#ifdef TEST_SECONDARY_SURFACE
    if( DirectDraw_SurfaceLock( secondary_surface ) )
        result = secondary_surface;
#else
    if( (!want_fullscreen) || !DirectDraw_SurfaceLock( primary_surface ) )
    {
        /* For windowed-mode Picasso screens, or for a failure in locking the primary-surface */
        /* lock the secondary surface, and draw to it.  A blit is then required from secondary to */
        /* primary surface. */
        if( DirectDraw_SurfaceLock( secondary_surface ) )
        {
            result = secondary_surface;
        }
    }
    else
    {
        result = primary_surface;
    }
#endif
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      Named this way for historical reasons
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
void unlockscr( void )
{
    lockcnt--;
    if( DirectDrawState.directdraw6 )
        IDirectDrawSurface4_Unlock( DirectDrawState.lockable.surface_dx6,
                                    DirectDrawState.lockable.lpdesc_dx6->lpSurface );
    else
        IDirectDrawSurface4_Unlock( DirectDrawState.lockable.surface,
                                    DirectDrawState.lockable.lpdesc->lpSurface );
}

#define disable_dx6() enable_directdraw6( __FILE__, __LINE__, 0 )
#define enable_dx6()  enable_directdraw6( __FILE__, __LINE__, 1 )

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
static void enable_directdraw6( const char *file, int line, int enable )
{
    DirectDrawState.directdraw6 = enable;
    if( !enable )
        write_log( "DirectDraw6 disabled [%s@%d] due to a failure condition!\n", file, line );
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
int start_ddraw (void)
{
    HRESULT ddrval;
    DDCAPS  drivercaps, helcaps;

    /* Prepare our DirectDrawState structure */
    ZeroMemory( &DirectDrawState, sizeof( DirectDrawState ) );
    if( currprefs.win32_directdraw6 )
        enable_dx6();

    ZeroMemory( &drivercaps, sizeof( drivercaps ) );
    ZeroMemory( &helcaps, sizeof( helcaps ) );
    drivercaps.dwSize = sizeof( DDCAPS );
    helcaps.dwSize = sizeof( DDCAPS );

    ddrval = (*pDirectDrawCreate)( NULL, &DirectDrawState.directdraw.dd, NULL );
    if (ddrval != DD_OK)
	goto oops;

    DirectDrawState.initialized = TRUE;

    ddrval = IDirectDraw_QueryInterface( DirectDrawState.directdraw.dd,
                                         &IID_IDirectDraw4,
                                         (LPVOID *)&DirectDrawState.directdraw.dd_dx6 );
    if( ddrval != DD_OK )
    {
        write_log( "System doesn't have DirectX-6 or higher... reverting to DirectX-3 style!\n" );
        disable_dx6();
    }

    DirectDraw_GetCaps( &drivercaps, &helcaps );
    ShowDDCaps( drivercaps, 1 );
    ShowDDCaps( helcaps, 0 );

    ddrval = DirectDraw_GetDisplayMode();
    if (ddrval != DD_OK)
	goto oops;

    return 1;

  oops:
    gui_message("start_ddraw(): DirectDraw initialization failed with %s/%d\n", DirectDraw_ErrorString (ddrval), ddrval);
    release_ddraw ();
    return 0;
}

#define releaser(x,y) if( x ) { y( x ); x = NULL; }

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
void release_ddraw (void)
{
    releaser( DirectDrawState.lpDDC, IDirectDrawClipper_Release );
    releaser( DirectDrawState.lpDDP, IDirectDrawPalette_Release );
    releaser( DirectDrawState.primary.surface_dx6, IDirectDrawSurface4_Release );
    releaser( DirectDrawState.primary.surface, IDirectDrawSurface_Release );
    releaser( DirectDrawState.secondary.surface_dx6, IDirectDrawSurface4_Release );
    releaser( DirectDrawState.secondary.surface, IDirectDrawSurface_Release );

    if( DirectDrawState.directdraw.dd_dx6 )
    {
	IDirectDraw4_RestoreDisplayMode( DirectDrawState.directdraw.dd_dx6 );
	IDirectDraw4_SetCooperativeLevel( DirectDrawState.directdraw.dd_dx6, hAmigaWnd, DDSCL_NORMAL);
    }
    releaser( DirectDrawState.directdraw.dd_dx6, IDirectDraw4_Release );

    if( DirectDrawState.directdraw.dd )
    {
	IDirectDraw_RestoreDisplayMode( DirectDrawState.directdraw.dd );
	IDirectDraw_SetCooperativeLevel( DirectDrawState.directdraw.dd, hAmigaWnd, DDSCL_NORMAL);
    }
    releaser( DirectDrawState.directdraw.dd, IDirectDraw_Release );

    DirectDrawState.lockable.lpdesc = NULL;
    DirectDrawState.lockable.lpdesc_dx6 = NULL;
    DirectDrawState.lockable.surface = NULL;
    DirectDrawState.lockable.surface_dx6 = NULL;

    DirectDrawState.surface_type = invalid_surface;
    DirectDrawState.initialized = FALSE;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_SetCooperativeLevel( HWND window, int want_fullscreen )
{
    HRESULT ddrval;

    if( DirectDrawState.directdraw6 )
    {
        ddrval = IDirectDraw4_SetCooperativeLevel( DirectDrawState.directdraw.dd_dx6,
                                                   window,
                                                   want_fullscreen ?
                                                   DDSCL_ALLOWREBOOT | DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN :
                                                   DDSCL_NORMAL );
    }
    else
    {
        ddrval = IDirectDraw_SetCooperativeLevel( DirectDrawState.directdraw.dd,
                                                   window,
                                                   want_fullscreen ?
                                                   DDSCL_ALLOWREBOOT | DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN :
                                                   DDSCL_NORMAL );
    }
    return ddrval;
}

/*
 * FUNCTION:   DirectDraw_SetDisplayMode
 *
 * PURPOSE:    Change the display-mode to width x height pixels, with a given
 *             vertical refresh-rate.
 *
 * PARAMETERS:
 *   width   - width of display in pixels
 *   height  - height of display in pixels
 *   freq    - vertical refresh-rate in Hz
 *
 * RETURNS:
 *   ddrval  - HRESULT indicating success (DD_OK) or failure
 *
 * NOTES:      The freq parameter is only obeyed on when we're using DirectX 6
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_SetDisplayMode( int width, int height, int bits, int freq )
{
    HRESULT ddrval;

    if( DirectDrawState.directdraw6 )
    {
        ddrval = IDirectDraw4_SetDisplayMode( DirectDrawState.directdraw.dd_dx6,
                                              width, height, bits, freq, 0 );
    }
    else
    {
        ddrval = IDirectDraw_SetDisplayMode( DirectDrawState.directdraw.dd,
                                              width, height, bits );
    }
    return ddrval;
}

/*
 * FUNCTION:   DirectDraw_GetDisplayMode
 *
 * PURPOSE:    Get the display-mode characteristics.
 *
 * PARAMETERS: none
 *
 * RETURNS:
 *   ddrval  - HRESULT indicating success (DD_OK) or failure
 *
 * NOTES:      none
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_GetDisplayMode( void )
{
    HRESULT ddrval;

    if( DirectDrawState.directdraw6 )
    {
        DirectDrawState.current.desc_dx6.dwSize = sizeof( DDSURFACEDESC2 );
        ddrval = IDirectDraw4_GetDisplayMode( DirectDrawState.directdraw.dd_dx6,
                                              &DirectDrawState.current.desc_dx6 );
    }

    /* We fill in the current.desc in all cases */
    DirectDrawState.current.desc.dwSize = sizeof( DDSURFACEDESC );
    ddrval = IDirectDraw_GetDisplayMode( DirectDrawState.directdraw.dd,
                                         &DirectDrawState.current.desc );
    return ddrval;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_GetCaps( DDCAPS *driver_caps, DDCAPS *hel_caps )
{
    HRESULT ddrval;

    if( DirectDrawState.directdraw6 )
    {
        ddrval = IDirectDraw4_GetCaps( DirectDrawState.directdraw.dd_dx6,
                                       driver_caps, hel_caps );
    }
    else
    {
        ddrval = IDirectDraw_GetCaps( DirectDrawState.directdraw.dd,
                                      driver_caps, hel_caps );
    }
    return ddrval;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_CreateClipper( void )
{
    HRESULT ddrval;
    if( DirectDrawState.directdraw6 )
    {
        ddrval = IDirectDraw4_CreateClipper( DirectDrawState.directdraw.dd_dx6,
                                             0,
                                             &DirectDrawState.lpDDC,
                                             NULL );
    }
    else
    {
        ddrval = IDirectDraw_CreateClipper( DirectDrawState.directdraw.dd,
                                             0,
                                             &DirectDrawState.lpDDC,
                                             NULL );
    }
    return ddrval;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_CreateSurface( surface_type_e type, int width, int height )
{
    HRESULT ddrval;

    if( type == primary_surface )
    {
        ZeroMemory( &DirectDrawState.primary.desc, sizeof( DDSURFACEDESC ) );
        DirectDrawState.primary.desc.dwSize = sizeof( DDSURFACEDESC );
        DirectDrawState.primary.desc.dwFlags = DDSD_CAPS;
        DirectDrawState.primary.desc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_VIDEOMEMORY | DDSCAPS_LOCALVIDMEM;
        ddrval = IDirectDraw_CreateSurface( DirectDrawState.directdraw.dd, 
                                            &DirectDrawState.primary.desc,
                                            &DirectDrawState.primary.surface,
                                            NULL );
        if( ddrval != DD_OK )
        {
            DirectDrawState.primary.desc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_VIDEOMEMORY | DDSCAPS_NONLOCALVIDMEM;
            ddrval = IDirectDraw_CreateSurface( DirectDrawState.directdraw.dd, 
                                                &DirectDrawState.primary.desc,
                                                &DirectDrawState.primary.surface,
                                                NULL );
            if( ddrval != DD_OK )
            {
                DirectDrawState.primary.desc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_SYSTEMMEMORY;
                ddrval = IDirectDraw_CreateSurface( DirectDrawState.directdraw.dd, 
                                                    &DirectDrawState.primary.desc,
                                                    &DirectDrawState.primary.surface,
                                                    NULL );
		if( ddrval == DD_OK )
		    write_log( "DDRAW: Primary surface created in plain system-memory\n" );
            }
            else
            {
                write_log( "DDRAW: Primary surface created in video system-memory\n" );
            }
        }
        else
        {
            write_log( "DDRAW: Primary surface created in video-memory\n" );
        }

        if( ddrval == DD_OK && DirectDrawState.directdraw6 )
        {
            ddrval = IDirectDrawSurface_QueryInterface( DirectDrawState.primary.surface,
                                                        &IID_IDirectDrawSurface4,
                                                        (LPVOID *)&DirectDrawState.primary.surface_dx6 );
            if( ddrval != DD_OK )
            {
                write_log( "System doesn't have DirectX-6 or higher... reverting to DirectX-1 style!\n" );
                disable_dx6();
                ddrval = DD_OK;
            }
        }
    }
    else
    {
        ZeroMemory( &DirectDrawState.secondary.desc, sizeof( DDSURFACEDESC ) );
        DirectDrawState.secondary.desc.dwSize = sizeof( DDSURFACEDESC );
        DirectDrawState.secondary.desc.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
        DirectDrawState.secondary.desc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_VIDEOMEMORY | DDSCAPS_LOCALVIDMEM;
        DirectDrawState.secondary.desc.dwWidth = width;
        DirectDrawState.secondary.desc.dwHeight = height;
        ddrval = IDirectDraw_CreateSurface( DirectDrawState.directdraw.dd, 
                                            &DirectDrawState.secondary.desc,
                                            &DirectDrawState.secondary.surface,
                                            NULL );
        if( ddrval != DD_OK )
        {
            DirectDrawState.secondary.desc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_VIDEOMEMORY | DDSCAPS_NONLOCALVIDMEM;
            ddrval = IDirectDraw_CreateSurface( DirectDrawState.directdraw.dd, 
                                                &DirectDrawState.secondary.desc,
                                                &DirectDrawState.secondary.surface,
                                                NULL );
            if( ddrval != DD_OK )
            {
                DirectDrawState.secondary.desc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
                ddrval = IDirectDraw_CreateSurface( DirectDrawState.directdraw.dd, 
                                                    &DirectDrawState.secondary.desc,
                                                    &DirectDrawState.secondary.surface,
                                                    NULL );
		if( ddrval == DD_OK )
		    write_log( "DDRAW: Secondary surface created in plain system-memory\n" );
            }
            else
            {
                write_log( "DDRAW: Secondary surface created in video system-memory\n" );
            }
        }
        else
        {
            write_log( "DDRAW: Secondary surface created in video-memory\n" );
        }

        if( ddrval == DD_OK && DirectDrawState.directdraw6 )
        {
            ddrval = IDirectDrawSurface_QueryInterface( DirectDrawState.secondary.surface,
                                                        &IID_IDirectDrawSurface4,
                                                        (LPVOID *)&DirectDrawState.secondary.surface_dx6 );
            if( ddrval != DD_OK )
            {
                write_log( "System doesn't have DirectX-6 or higher... reverting to DirectX-1 style!\n" );
                disable_dx6();
                ddrval = DD_OK;
            }
        }
    }
    return ddrval;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
int DirectDraw_DetermineLocking( int wantfull )
{
    int result = 0;

    switch( DirectDrawState.surface_type = try_surface_locks( wantfull ) )
    {
    case invalid_surface:
    case lockable_surface:
        DirectDrawState.lockable.lpdesc = NULL;
        DirectDrawState.lockable.lpdesc_dx6 = NULL;
        DirectDrawState.lockable.surface = NULL;
        DirectDrawState.lockable.surface_dx6 = NULL;
        write_log( "set_ddraw: Couldn't lock primary, and no secondary available.\n" );
        break;
    case primary_surface:
        DirectDrawState.lockable.lpdesc = &DirectDrawState.primary.desc;
        DirectDrawState.lockable.lpdesc_dx6 = &DirectDrawState.primary.desc_dx6;
        DirectDrawState.lockable.surface = DirectDrawState.primary.surface;
        DirectDrawState.lockable.surface_dx6 = DirectDrawState.primary.surface_dx6;
	result = 1;
        break;
    case secondary_surface:
        DirectDrawState.lockable.lpdesc = &DirectDrawState.secondary.desc;
        DirectDrawState.lockable.lpdesc_dx6 = &DirectDrawState.secondary.desc_dx6;
        DirectDrawState.lockable.surface = DirectDrawState.secondary.surface;
        DirectDrawState.lockable.surface_dx6 = DirectDrawState.secondary.surface_dx6;

        write_log( "set_ddraw: Couldn't lock primary, so offscreen secondary buffer is being used.\n" );
	result = 1;
        break;
    }

    if( DirectDrawState.lockable.surface )
        DirectDraw_SurfaceUnlock();

    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_SetClipper( HWND hWnd )
{
    HRESULT ddrval;

    if( DirectDrawState.directdraw6 )
    {
        ddrval = IDirectDrawSurface4_SetClipper( DirectDrawState.primary.surface_dx6,
                                                 hWnd ? DirectDrawState.lpDDC : NULL );
    }
    else
    {
        ddrval = IDirectDrawSurface_SetClipper( DirectDrawState.primary.surface,
                                                hWnd ? DirectDrawState.lpDDC : NULL );
    }
    if( hWnd && ( ddrval == DD_OK ) )
    {
        ddrval = IDirectDrawClipper_SetHWnd( DirectDrawState.lpDDC, 0, hWnd );
    }
    return ddrval;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_GetClipList( LPRGNDATA cliplist, LPDWORD size )
{
    HRESULT ddrval;

    ddrval = IDirectDrawClipper_GetClipList( DirectDrawState.lpDDC, NULL, cliplist, size );

    return ddrval;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
BYTE DirectDraw_GetBytesPerPixel( void )
{
    int bpp;
    if( DirectDrawState.directdraw6 )
    {
        bpp = ( DirectDrawState.lockable.lpdesc_dx6->ddpfPixelFormat.dwRGBBitCount + 7 ) >> 3;
    }
    else
    {
        bpp = ( DirectDrawState.lockable.lpdesc->ddpfPixelFormat.dwRGBBitCount + 7 ) >> 3;
    }
    return bpp;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_SetPalette( int remove )
{
    HRESULT ddrval;
    if( DirectDrawState.directdraw6 )
    {
        ddrval = IDirectDrawSurface4_SetPalette( DirectDrawState.primary.surface_dx6,
                                                 remove ? NULL : DirectDrawState.lpDDP );
    }
    else
    {
        ddrval = IDirectDrawSurface_SetPalette( DirectDrawState.primary.surface,
                                                remove ? NULL : DirectDrawState.lpDDP );
    }
    return ddrval;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_CreatePalette( LPPALETTEENTRY pal )
{
    HRESULT ddrval;
    if( DirectDrawState.directdraw6 )
    {
        ddrval = IDirectDraw4_CreatePalette( DirectDrawState.directdraw.dd_dx6,
                                            DDPCAPS_8BIT | DDPCAPS_ALLOW256,
                                            pal,
                                            &DirectDrawState.lpDDP, NULL);
    }
    else
    {
        ddrval = IDirectDraw_CreatePalette( DirectDrawState.directdraw.dd,
                                            DDPCAPS_8BIT | DDPCAPS_ALLOW256,
                                            pal,
                                            &DirectDrawState.lpDDP, NULL);
    }
    if( ddrval == DD_OK )
    {
        ddrval = DirectDraw_SetPalette(0);
    }
    return ddrval;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_SetPaletteEntries( int start, int count, PALETTEENTRY *palette )
{
    HRESULT ddrval = DDERR_NOPALETTEATTACHED;
    if( DirectDrawState.lpDDP )
        ddrval = IDirectDrawPalette_SetEntries( DirectDrawState.lpDDP, 0, start, count, palette );
    return ddrval;
}

/* Return one of the pixel formats declared in picasso96.h if the surface
 * is usable for us, or RGBFB_NONE if it is not usable.  */
/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
RGBFTYPE DirectDraw_GetSurfacePixelFormat( LPVOID surface )
{
    int surface_is_dx6 = 0;
    DDPIXELFORMAT *pfp = NULL;
    DWORD r, g, b;

    /* These are all the surface-flags that we require */
    DWORD flags = (DDSD_PIXELFORMAT | DDSD_WIDTH | DDSD_HEIGHT | DDSD_REFRESHRATE);
    DWORD surf_flags;
    
    LPDDSURFACEDESC lpsd = NULL;
    LPDDSURFACEDESC2 lpsd2 = NULL;

    /* Examine the size to see what type of surface-descriptor this is */
    if( ((LPDDSURFACEDESC)surface)->dwSize == sizeof( DDSURFACEDESC ) )
    {
        lpsd = (LPDDSURFACEDESC)surface;
        surf_flags = lpsd->dwFlags;
        pfp = &lpsd->ddpfPixelFormat;
    }
    else
    {
        lpsd2 = (LPDDSURFACEDESC2)surface;
        surf_flags = lpsd2->dwFlags;
        pfp = &lpsd2->ddpfPixelFormat;
    }

    if( ( surf_flags & flags ) != flags )
	return RGBFB_NONE;

    if ((pfp->dwFlags & DDPF_RGB) == 0)
	return RGBFB_NONE;

    r = pfp->dwRBitMask;
    g = pfp->dwGBitMask;
    b = pfp->dwBBitMask;
    switch (pfp->dwRGBBitCount) {
     case 8:
	if ((pfp->dwFlags & DDPF_PALETTEINDEXED8) != 0)
	    return RGBFB_CHUNKY;
	break;

     case 16:
	if (r == 0xF800 && g == 0x07E0 && b == 0x001F)
	    return RGBFB_R5G6B5PC;
	if (r == 0x7C00 && g == 0x03E0 && b == 0x001F)
	    return RGBFB_R5G5B5PC;
	if (b == 0xF800 && g == 0x07E0 && r == 0x001F)
	    return RGBFB_B5G6R5PC;
	if (b == 0x7C00 && g == 0x03E0 && r == 0x001F)
	    return RGBFB_B5G5R5PC;
	/* This happens under NT - with r == b == g == 0 !!! */
	write_log ("Unknown 16 bit format %d %d %d\n", r, g, b);
	break;

     case 24:
	if (r == 0xFF0000 && g == 0x00FF00 && b == 0x0000FF)
	    return RGBFB_B8G8R8;
	if (r == 0x0000FF && g == 0x00FF00 && b == 0xFF0000)
	    return RGBFB_R8G8B8;
	break;

     case 32:
	if (r == 0x00FF0000 && g == 0x0000FF00 && b == 0x000000FF)
	    return RGBFB_B8G8R8A8;
	if (r == 0x000000FF && g == 0x0000FF00 && b == 0x00FF0000)
	    return RGBFB_R8G8B8A8;
	if (r == 0xFF000000 && g == 0x00FF0000 && b == 0x0000FF00)
	    return RGBFB_A8B8G8R8;
	if (r == 0x0000FF00 && g == 0x00FF0000 && b == 0xFF000000)
	    return RGBFB_A8R8G8B8;
	break;
	
     default:
	gui_message("Unknown %d bit format %d %d %d\n", pfp->dwRGBBitCount, r, g, b); /* %%% - BERND, and here too... */
	break;
    }
    return RGBFB_NONE;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
RGBFTYPE DirectDraw_GetPixelFormat( void )
{
    RGBFTYPE type;
    if( DirectDrawState.directdraw6 )
    {
        type = DirectDraw_GetSurfacePixelFormat( &DirectDrawState.current.desc_dx6 );
    }
    else
    {
        type = DirectDraw_GetSurfacePixelFormat( &DirectDrawState.current.desc );
    }
    return type;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
DWORD DirectDraw_CurrentWidth( void )
{
    DWORD width;
    if( DirectDrawState.directdraw6 )
    {
        width = DirectDrawState.current.desc_dx6.dwWidth;
    }
    else
    {
        width = DirectDrawState.current.desc.dwWidth;
    }
    return width;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
DWORD DirectDraw_CurrentHeight( void )
{
    DWORD height;
    if( DirectDrawState.directdraw6 )
    {
        height = DirectDrawState.current.desc_dx6.dwHeight;
    }
    else
    {
        height = DirectDrawState.current.desc.dwHeight;
    }
    return height;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
static int DirectDraw_BltFastStub( LPDIRECTDRAWSURFACE dstsurf, DWORD x, DWORD y, LPDIRECTDRAWSURFACE srcsurf, LPRECT srcrect )
{
    int result = 0;
    HRESULT ddrval;

    while( ( ddrval = IDirectDrawSurface_BltFast( dstsurf, x, y, srcsurf, srcrect, DDBLTFAST_NOCOLORKEY | DDBLTFAST_WAIT ) ) != DD_OK )
    {
        if (ddrval == DDERR_SURFACELOST) 
        {
    	    ddrval = IDirectDrawSurface_Restore( dstsurf );
            if (ddrval != DD_OK)
            {
                break;
            }
        }
        else if (ddrval != DDERR_SURFACEBUSY) 
        {
	    write_log("BltFastStub(): DirectDrawSurface_BltFast() failed with %s/%d\n", DirectDraw_ErrorString (ddrval), ddrval);
            break;
        }
    }
    if( ddrval == DD_OK )
    {
        result = 1;
    }
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
static int DirectDraw_BltFastStub4( LPDIRECTDRAWSURFACE4 dstsurf, DWORD x, DWORD y, LPDIRECTDRAWSURFACE4 srcsurf, LPRECT srcrect )
{
    int result = 0;
    HRESULT ddrval;

    while( ( ddrval = IDirectDrawSurface4_BltFast( dstsurf, x, y, srcsurf, srcrect, DDBLTFAST_NOCOLORKEY | DDBLTFAST_WAIT ) ) != DD_OK )
    {
        if (ddrval == DDERR_SURFACELOST) 
        {
    	    ddrval = IDirectDrawSurface4_Restore( dstsurf );
            if (ddrval != DD_OK)
            {
                break;
            }
        }
        else if (ddrval != DDERR_SURFACEBUSY) 
        {
	    write_log("BltFastStub4(): DirectDrawSurface4_BltFast() failed with %s/%d\n", DirectDraw_ErrorString (ddrval), ddrval);
            break;
        }
    }
    if( ddrval == DD_OK )
    {
        result = 1;
    }
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
int DirectDraw_BltFast( surface_type_e dsttype, DWORD left, DWORD top, surface_type_e srctype, LPRECT srcrect )
{
    int result;

    if( DirectDrawState.directdraw6 )
    {
        LPDIRECTDRAWSURFACE4 lpDDS4_dst, lpDDS4_src;
        if( dsttype == primary_surface )
        {
            lpDDS4_dst = DirectDrawState.primary.surface_dx6;
        }
        else
        {
            lpDDS4_dst = DirectDrawState.secondary.surface_dx6;
        }
        if( srctype == primary_surface )
        {
            lpDDS4_src = DirectDrawState.primary.surface_dx6;
        }
        else
        {
            lpDDS4_src = DirectDrawState.secondary.surface_dx6;
        }
        result = DirectDraw_BltFastStub4( lpDDS4_dst, left, top, lpDDS4_src, srcrect );
    }
    else
    {
        LPDIRECTDRAWSURFACE lpDDS_dst, lpDDS_src;
        if( dsttype == primary_surface )
        {
            lpDDS_dst = DirectDrawState.primary.surface;
        }
        else
        {
            lpDDS_dst = DirectDrawState.secondary.surface;
        }
        if( srctype == primary_surface )
        {
            lpDDS_src = DirectDrawState.primary.surface;
        }
        else
        {
            lpDDS_src = DirectDrawState.secondary.surface;
        }
        result = DirectDraw_BltFastStub( lpDDS_dst, left, top, lpDDS_src, srcrect );
    }
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
static int DirectDraw_BltStub( LPDIRECTDRAWSURFACE dstsurf, LPRECT dstrect, LPDIRECTDRAWSURFACE srcsurf, LPRECT srcrect, DWORD flags, LPDDBLTFX ddbltfx )
{
    int result = 0;
    HRESULT ddrval;

    while( ( ddrval = IDirectDrawSurface_Blt( dstsurf, dstrect, srcsurf, srcrect, flags, ddbltfx ) ) != DD_OK )
    {
        if (ddrval == DDERR_SURFACELOST) 
        {
    	    ddrval = IDirectDrawSurface_Restore( dstsurf );
            if (ddrval != DD_OK)
            {
                break;
            }
        }
        else if (ddrval != DDERR_SURFACEBUSY) 
        {
	    write_log("BltStub(): DirectDrawSurface_Blt() failed with %s/%d\n", DirectDraw_ErrorString (ddrval), ddrval);
            break;
        }
    }
    if( ddrval == DD_OK )
    {
        result = 1;
    }
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
static int DirectDraw_BltStub4( LPDIRECTDRAWSURFACE4 dstsurf, LPRECT dstrect, LPDIRECTDRAWSURFACE4 srcsurf, LPRECT srcrect, DWORD flags, LPDDBLTFX ddbltfx )
{
    int result = 0;
    HRESULT ddrval;

    while( ( ddrval = IDirectDrawSurface4_Blt( dstsurf, dstrect, srcsurf, srcrect, flags, ddbltfx ) ) != DD_OK )
    {
        if (ddrval == DDERR_SURFACELOST) 
        {
    	    ddrval = IDirectDrawSurface4_Restore( dstsurf );
            if (ddrval != DD_OK)
            {
                break;
            }
        }
        else if (ddrval != DDERR_SURFACEBUSY) 
        {
	    write_log("BltStub4(): DirectDrawSurface4_Blt() failed with %s/%d\n", DirectDraw_ErrorString (ddrval), ddrval);
            break;
        }
    }
    if( ddrval == DD_OK )
    {
        result = 1;
    }
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
int DirectDraw_Blt( surface_type_e dsttype, LPRECT dstrect,
                    surface_type_e srctype, LPRECT srcrect,
                    DWORD flags, LPDDBLTFX fx )
{
    int result;

    if( DirectDrawState.directdraw6 )
    {
        LPDIRECTDRAWSURFACE4 lpDDS4_dst, lpDDS4_src;
        if( dsttype == primary_surface )
        {
            lpDDS4_dst = DirectDrawState.primary.surface_dx6;
        }
        else
        {
            lpDDS4_dst = DirectDrawState.secondary.surface_dx6;
        }
        if( srctype == primary_surface )
        {
            lpDDS4_src = DirectDrawState.primary.surface_dx6;
        }
        else if( srctype == secondary_surface )
        {
            lpDDS4_src = DirectDrawState.secondary.surface_dx6;
        }
	else
	{
	    lpDDS4_src = NULL; /* For using BltStub to do rect-fills */
	}
        result = DirectDraw_BltStub4( lpDDS4_dst, dstrect, lpDDS4_src, srcrect, flags, fx );
    }
    else
    {
        LPDIRECTDRAWSURFACE lpDDS_dst, lpDDS_src;
        if( dsttype == primary_surface )
        {
            lpDDS_dst = DirectDrawState.primary.surface;
        }
        else
        {
            lpDDS_dst = DirectDrawState.secondary.surface;
        }
        if( srctype == primary_surface )
        {
            lpDDS_src = DirectDrawState.primary.surface;
        }
        else if( srctype == secondary_surface )
        {
            lpDDS_src = DirectDrawState.secondary.surface;
        }
        else
        {
            lpDDS_src = NULL; /* For using BltStub to do rect-fills */
        }
        result = DirectDraw_BltStub( lpDDS_dst, dstrect, lpDDS_src, srcrect, flags, fx );
    }
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_WaitForVerticalBlank( DWORD flags )
{
    HRESULT result;
    if( DirectDrawState.directdraw6 )
    {
        result = IDirectDraw4_WaitForVerticalBlank( DirectDrawState.directdraw.dd_dx6, flags, NULL );
    }
    else
    {
        result = IDirectDraw_WaitForVerticalBlank( DirectDrawState.directdraw.dd, flags, NULL );
    }
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_EnumDisplayModes( DWORD flags, LPDDENUMMODESCALLBACK callback, LPDDENUMMODESCALLBACK2 callback2 )
{
    HRESULT result;
    if( DirectDrawState.directdraw6 )
    {
        result = IDirectDraw4_EnumDisplayModes( DirectDrawState.directdraw.dd_dx6, flags, NULL, NULL, callback2 );    
    }
    else
    {
        result = IDirectDraw_EnumDisplayModes( DirectDrawState.directdraw.dd, flags, NULL, NULL, callback );    
    }
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_FlipToGDISurface( void )
{
    HRESULT result = DDERR_GENERIC;
    if( DirectDrawState.initialized )
    {
	if( DirectDrawState.directdraw6 )
	{
	    result = IDirectDraw4_FlipToGDISurface( DirectDrawState.directdraw.dd_dx6 );
	}
	else
	{
	    result = IDirectDraw_FlipToGDISurface( DirectDrawState.directdraw.dd );
	}
    }
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_GetDC( HDC *hdc, surface_type_e surface )
{
    HRESULT result;
    if( DirectDrawState.directdraw6 )
    {
        if( surface == primary_surface )
            result = IDirectDrawSurface4_GetDC( DirectDrawState.primary.surface_dx6, hdc );
        else
            result = IDirectDrawSurface4_GetDC( DirectDrawState.secondary.surface_dx6, hdc );
    }
    else
    {
        if( surface == primary_surface )
            result = IDirectDrawSurface_GetDC( DirectDrawState.primary.surface, hdc );
        else
            result = IDirectDrawSurface_GetDC( DirectDrawState.secondary.surface, hdc );
    }
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_ReleaseDC( HDC hdc, surface_type_e surface )
{
    HRESULT result;
    if( DirectDrawState.directdraw6 )
    {
        if( surface == primary_surface )
            result = IDirectDrawSurface4_ReleaseDC( DirectDrawState.primary.surface_dx6, hdc );
        else
            result = IDirectDrawSurface4_ReleaseDC( DirectDrawState.secondary.surface_dx6, hdc );
    }
    else
    {
        if( surface == primary_surface )
            result = IDirectDrawSurface_ReleaseDC( DirectDrawState.primary.surface, hdc );
        else
            result = IDirectDrawSurface_ReleaseDC( DirectDrawState.secondary.surface, hdc );
    }
    return result;
}
