/*
 * UAE - The Un*x Amiga Emulator
 *
 * Win32 interface
 *
 * Copyright 1997-1998 Mathias Ortmann
 * Copyright 1997-1999 Brian King
 */

/* Uncomment this line if you want the logs time-stamped */
/* #define TIMESTAMP_LOGS */

#include "config.h"
#include "sysconfig.h"

#include <stdlib.h>
#include <stdarg.h>
#include <signal.h>

#include <windows.h>
#include <commctrl.h>
#include <commdlg.h>
#include <shellapi.h>
#ifdef _MSC_VER
#include <zmouse.h>
#include <ddraw.h>
#include <htmlhelp.h>
#else
#include "winstuff.h"
#endif
#include <io.h>

#include "sysdeps.h"
#include "options.h"
#include "gensound.h"
#include "sounddep/sound.h"
#include "uae.h"
#include "memory.h"
#include "custom.h"
#include "events.h"
#include "xwin.h"
#include "keyboard.h"
#include "keybuf.h"
#include "drawing.h"
#include "picasso96.h"
#include "bsdsocket.h"
#include "osdep/win32.h"
#include "osdep/win32gui.h"
#include "osdep/win32gfx.h"
#include "osdep/dxwrap.h"
#include "resource.h"
#include "gui.h"
#include "machdep/maccess.h"
#include "newcpu.h"

int useqpc = 0; /* Set to TRUE to use the QueryPerformanceCounter() function instead of rdtsc() */

static FILE *debugfile = NULL;

HINSTANCE hInst;

HWND (WINAPI *pHtmlHelp)(HWND, LPCSTR, UINT, LPDWORD ) = NULL;

HWND hAmigaWnd, hMainWnd;
/*DWORD Keys; */
RECT amigawin_rect;

char VersionStr[256];

int in_sizemove = 0;
int manual_painting_needed = 0;
int customsize = 0;

int bActive;
int toggle_sound;

BOOL viewing_child = FALSE;

HKEY hWinUAEKey    = NULL;

/* Keyboard emulation, Win32 helper routines. */
static LPARAM keysdown[256];
static short numkeysdown;
int checkkey (int vkey, LPARAM lParam)
{
    switch (vkey) {
    case VK_LWIN:
    case VK_RWIN:
     case VK_SHIFT:
     case VK_LSHIFT:
     case VK_RSHIFT:
     case VK_CONTROL:
     case VK_LCONTROL:
     case VK_RCONTROL:
     case VK_MENU:
     case VK_LMENU:
     case VK_RMENU:
	return GetKeyState (vkey) & 0x8000;
    }
    return GetAsyncKeyState (vkey) & 0x8000;
}

/* Mouse emulation, Win32 interface */
static int mousecx = 160, mousecy = 100, mousedx = 160, mousedy = 100;
static int mousecl = MAKELONG (160, 100);
int mouseactive;

void WIN32_MouseDefaults( void )
{
    mousecx = 160, mousecy = 100, mousedx = 160, mousedy = 100, mousecl = MAKELONG (160, 100);
}

void setmouseactive (int active)
{
    mousedx = (amigawin_rect.right - amigawin_rect.left) / 2;
    mousedy = (amigawin_rect.bottom - amigawin_rect.top) / 2;
    mousecl = MAKELONG (mousedx, mousedy);
    mousecx = amigawin_rect.left + mousedx;
    mousecy = amigawin_rect.top + mousedy;

    if (active == mouseactive)
	return;
    mouseactive = active;

    if (active) {
	ShowCursor (FALSE);
	SetCursorPos (mousecx, mousecy);
    if( currprefs.win32_middle_mouse )
    	SetWindowText (hMainWnd ? hMainWnd : hAmigaWnd, "WinUAE - [Mouse active - press Alt-Tab or middle-button to cancel]");
    else
    	SetWindowText (hMainWnd ? hMainWnd : hAmigaWnd, "WinUAE - [Mouse active - press Alt-Tab to cancel]");
	ClipCursor (&amigawin_rect);
    } else {
	ShowCursor (TRUE);
	SetWindowText (hMainWnd ? hMainWnd : hAmigaWnd, "WinUAE" );
	ClipCursor (NULL);
    }
}

static int hascapture = 0;

static void setcapture (void)
{
    if (hascapture)
	return;
    hascapture++;
    SetCapture (hAmigaWnd);
}

void releasecapture (void)
{
    if (!hascapture)
	return;
    hascapture--;
    ReleaseCapture ();
}

frame_time_t read_processor_time_cyrix (void)
{
    LARGE_INTEGER counter;
    QueryPerformanceCounter( &counter );
    return (frame_time_t)(counter.u.LowPart);
}

static void figure_processor_speed (void)
{
    LARGE_INTEGER perfreq = {0,0};
    extern frame_time_t syncbase, vsynctime;
    frame_time_t clockrate;

#if defined( __GNUC__ ) && defined( X86_ASSEMBLY )
    extern int rdtsc_check( void ) __asm__ ("rdtsc_check");
    if( !rdtsc_check() )
    {
        useqpc = 1;
    }
#endif
    rpt_available = 1; /* ALWAYS available on Win32, because we've either got RDTSC or QueryPerformanceCounter() */

    write_log( "Measuring Clock Rate... " );
    SetPriorityClass( GetCurrentProcess(), REALTIME_PRIORITY_CLASS );
    clockrate = read_processor_time();
    Sleep( 1000 );
    clockrate = read_processor_time() - clockrate;
    SetPriorityClass( GetCurrentProcess(), NORMAL_PRIORITY_CLASS );

    write_log( "%dMHz\n", clockrate / 1000000 );
    syncbase = clockrate;
    vsynctime = syncbase / VBLANK_HZ_PAL; /* default to 50Hz */
}

static long FAR PASCAL AmigaWindowProc (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    PAINTSTRUCT ps;
    HDC hDC;
    short wheeldelta;
    static int store_xy = 0;
    BOOL minimized;

    switch( message ) 
    {
        case WM_PALETTECHANGED:
            if( (HWND)wParam != hWnd )
            {
                write_log( "WM_PALETTECHANGED Request\n" );
                WIN32GFX_PaletteChange();
            }
        break;

        case WM_ACTIVATEAPP:
            if (bActive = wParam) 
            {
		if( WIN32GFX_IsFullScreen() ) 
                {
		    SetCursor (NULL);
		    SetCursorPos (mousecx, mousecy);
	        }
    	        my_kbd_handler (VK_CAPITAL, 0x3a, TRUE);
	    }
            else
            {
                if( !WIN32GFX_IsFullScreen() )
		    setmouseactive (0);
                else
                {
                    if( WIN32GFX_IsPicassoScreen() )
                    {
                        WIN32GFX_DisablePicasso();
                    }
                }
	        }
	    break;

        case WM_ACTIVATE:
            minimized = HIWORD( wParam );
            if (LOWORD (wParam) != WA_INACTIVE) 
            {
                write_log( "WinUAE now active via WM_ACTIVATE\n" );
                if( !minimized )
                {
                    if( currprefs.win32_iconified_nospeed )
                    {
                        SetPriorityClass( GetCurrentProcess(), NORMAL_PRIORITY_CLASS );
                    }
                    if( currprefs.win32_iconified_nosound )
                    {
                        resume_sound();
                    }
                }
	        ShowWindow (hWnd, SW_RESTORE);
                if( WIN32GFX_IsPicassoScreen() )
                {
                    WIN32GFX_EnablePicasso();
                }
	    }
            else
            {
                write_log( "WinUAE now inactive via WM_ACTIVATE\n" );
                if( minimized && !quit_program )
                {
                    if( currprefs.win32_iconified_nospeed )
                    {
                        SetPriorityClass( GetCurrentProcess(), IDLE_PRIORITY_CLASS );
                    }
                    if( currprefs.win32_iconified_nosound )
                    {
                        pause_sound(1);
                    }
                }
            }
	    break;

        case WM_SETCURSOR:
	        if( WIN32GFX_IsFullScreen() ) 
            {
	            SetCursor (NULL);
	            return TRUE;
	        }
	    break;

        case WM_SYSCOMMAND:
            switch( wParam & 0xFFF0 )
            {
                case SC_MAXIMIZE:
                    WIN32GFX_ToggleFullScreen();
	                return 0;
                break;
                case SC_MINIMIZE:
                    WIN32GFX_DisablePicasso();
                break;
            }
	    break;

        case WM_KEYUP:
        case WM_SYSKEYUP:
	        numkeysdown--;
	        keysdown[wParam] = 0;
	        my_kbd_handler (wParam, (lParam >> 16) & 0x1ff, FALSE);
            return 0;
    	break;

     case WM_KEYDOWN:
     case WM_SYSKEYDOWN:
	if (LOWORD (lParam) == 1) 
    {
	    if (numkeysdown) 
        {
		    int key;
		    numkeysdown = 0;

		    for (key = 256; key--;) 
            {
		        if (keysdown[key]) 
                {
			        if (checkkey (key, lParam))
			            numkeysdown++;
			        else 
                    {
			            my_kbd_handler (key, (keysdown[key] >> 16) & 0x1ff, FALSE);
			            keysdown[key] = 0;
			        }
		        }
		    }
	    }
	    if (!keysdown[wParam]) 
        {
		    keysdown[wParam] = lParam;
		    numkeysdown++;
	    }
	    numkeysdown++;
	    my_kbd_handler (wParam, (lParam >> 16) & 0x1ff, TRUE);
	}
	break;

     case WM_LBUTTONDOWN:
	if (ievent_alive) {
	    setcapture ();
	    buttonstate[0] = 1;
	} else if (!WIN32GFX_IsFullScreen() && !mouseactive)
	    setmouseactive (1);
	else
	    buttonstate[0] = 1;
	break;

     case WM_LBUTTONUP:
	releasecapture ();
	buttonstate[0] = 0;
	break;

     case WM_MBUTTONDOWN:
	if (ievent_alive)
	    setcapture ();
	buttonstate[1] = 1;
	break;

     case WM_MBUTTONUP:
	releasecapture ();
	buttonstate[1] = 0;
    /* Do some fun stuff to handle Julian's request for the middle-mouse-button to act like ALT-TAB */
    if( !WIN32GFX_IsFullScreen() && mouseactive && currprefs.win32_middle_mouse )
        setmouseactive(0);
	break;

     case WM_RBUTTONDOWN:
	if (ievent_alive)
	    setcapture ();
	buttonstate[2] = 1;
	break;

     case WM_RBUTTONUP:
	releasecapture ();
	buttonstate[2] = 0;
	break;

     case WM_VSCROLL:
         write_log( "WM_VSCROLL\n" );
         if( LOWORD( wParam ) == SB_LINEDOWN )
             record_key(0x7A<<1);
         else if( LOWORD( wParam ) == SB_LINEUP )
             record_key(0x7B<<1);
         break;

     case WM_MOUSEWHEEL:
         write_log( "WM_MOUSEWHEEL\n" );
         wheeldelta = HIWORD(wParam);
         if( wheeldelta > 0 )
             record_key(0x7A<<1);
         else
             record_key(0x7B<<1);
         break;

     case WM_MOUSEMOVE:
	if( ( mouseactive && !ievent_alive ) || 
        WIN32GFX_IsFullScreen() ) 
    {
	    /*
	     * In this mode, the mouse pointer is always centered in the window,
	     * this is ensured by the SetCursorPos call below.
	     * We don't want to handle messages that result from such a SetCursorPos
	     * call (recursion!), so exit early if we see one.
	     */
	    if (lParam == mousecl)
		break;
	    lastmx += (signed short) LOWORD (lParam) - mousedx;
	    lastmy += (signed short) HIWORD (lParam) - mousedy;
	    if (ievent_alive) {
		if (lastmx < 0)
		    lastmx = 0;
		if (lastmx > WIN32GFX_GetWidth() )
		    lastmx = WIN32GFX_GetWidth();
		if (lastmy < 0)
		    lastmy = 0;
		if (lastmy > WIN32GFX_GetHeight() )
		    lastmy = WIN32GFX_GetHeight();
	    }
	    SetCursorPos (mousecx, mousecy);
	    break;
	}
	lastmx = (signed short) LOWORD (lParam);
	lastmy = (signed short) HIWORD (lParam);
	break;

     case WM_PAINT:
	clear_inhibit_frame (IHF_WINDOWHIDDEN);
	notice_screen_contents_lost ();

        hDC = BeginPaint (hWnd, &ps);

        /* Check to see if this WM_PAINT is coming while we've got the GUI visible */
        if( manual_painting_needed )
        {
            /* Update the display area */
            if( !WIN32GFX_IsFullScreen() )
            {
                flush_screen( 0, WIN32GFX_GetHeight()-1 );
            }
            else
            {
                DirectDraw_Blt( primary_surface, NULL, secondary_surface, NULL, DDBLT_WAIT, NULL );
            }
        }

	EndPaint (hWnd, &ps);

	break;

     case WM_DROPFILES:
	if (DragQueryFile ((HDROP) wParam, (UINT) - 1, NULL, 0)) {
	    if (DragQueryFile ((HDROP) wParam, 0, NULL, 0) < 255)
		DragQueryFile ((HDROP) wParam, 0, changed_prefs.df[0], sizeof (changed_prefs.df[0]));
	}
	DragFinish ((HDROP) wParam);
	break;

     case WM_CAPTURECHANGED:
	if ((HWND)lParam != hAmigaWnd)
	    buttonstate[0] = buttonstate[1] = buttonstate[2] = 0;
	break;

     case WM_TIMER:
	finishjob ();
	break;

     case WM_CREATE:
	DragAcceptFiles (hWnd, TRUE);
	break;

     case WM_CLOSE:
	uae_quit ();
	return 0;

     case WM_WINDOWPOSCHANGED:
	if( GetWindowRect(hAmigaWnd, &amigawin_rect) )
	{
	    if( hMainWnd )
	    {
		if( hWinUAEKey && store_xy++)
		{
		    RegSetValueEx( hWinUAEKey, "xPos", 0, REG_DWORD, (LPBYTE)&amigawin_rect.left, sizeof( LONG ) );
		    RegSetValueEx( hWinUAEKey, "yPos", 0, REG_DWORD, (LPBYTE)&amigawin_rect.top, sizeof( LONG ) );
		}
	    }
	}
	break;

    default:
#if 0
#ifdef BSDSOCKET_SUPPORTED
    if( message >= 0xB000 && message < 0xB000+MAXPENDINGASYNC*2 )
    {
#if DEBUG_SOCKETS
        write_log( "sockmsg(0x%x, 0x%x, 0x%x)\n", message, wParam, lParam );
#endif
        sockmsg( message, wParam, lParam );
        return 0;
    }
#endif
#endif
    break;
    }

    return DefWindowProc (hWnd, message, wParam, lParam);
}

static long FAR PASCAL MainWindowProc (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    PAINTSTRUCT ps;
    RECT rc;
    HDC hDC;

    switch (message) {
     case WM_LBUTTONDOWN:
     case WM_MOUSEWHEEL:
     case WM_MOUSEMOVE:
     case WM_ACTIVATEAPP:
     case WM_ACTIVATE:
     case WM_SETCURSOR:
     case WM_SYSCOMMAND:
     case WM_KEYUP:
     case WM_SYSKEYUP:
     case WM_KEYDOWN:
     case WM_SYSKEYDOWN:
     case WM_LBUTTONUP:
     case WM_MBUTTONDOWN:
     case WM_MBUTTONUP:
     case WM_RBUTTONDOWN:
     case WM_RBUTTONUP:
     case WM_DROPFILES:
     case WM_CREATE:
     case WM_DESTROY:
     case WM_USER + 0x200:
     case WM_CLOSE:
     case WM_HELP:
	return AmigaWindowProc (hWnd, message, wParam, lParam);

     case WM_DISPLAYCHANGE:
	if (!WIN32GFX_IsFullScreen && (wParam + 7) / 8 != DirectDraw_GetBytesPerPixel() )
	    WIN32GFX_DisplayChangeRequested();
	break;
	
     case WM_ENTERSIZEMOVE:
	in_sizemove++;
	break;

     case WM_EXITSIZEMOVE:
	in_sizemove--;

	/* fall through */

     case WM_WINDOWPOSCHANGED:
	if( GetWindowRect (hAmigaWnd, &amigawin_rect) )
	{
	    if (in_sizemove > 0)
		break;

	    if( !WIN32GFX_IsFullScreen() && hAmigaWnd )
	    {
		if( amigawin_rect.left & 3 ) 
		{
		    RECT rc2;
		    if( GetWindowRect( hMainWnd, &rc2 ) )
		    {
			MoveWindow (hMainWnd, rc2.left + 4 - amigawin_rect.left % 4, rc2.top,
				    rc2.right - rc2.left, rc2.bottom - rc2.top, TRUE);
		    }
		}
		//setmouseactive (0);
		return 0;
	    }
	}
	break;

     case WM_PAINT:
	hDC = BeginPaint (hWnd, &ps);
	GetClientRect (hWnd, &rc);
	DrawEdge (hDC, &rc, EDGE_SUNKEN, BF_RECT);
	EndPaint (hWnd, &ps);
	break;

     case WM_NCLBUTTONDBLCLK:
	if (wParam == HTCAPTION) {
	    WIN32GFX_ToggleFullScreen();
	    return 0;
	}
	break;

    default:
    break;

    }

    return DefWindowProc (hWnd, message, wParam, lParam);
}

void handle_events (void)
{
    MSG msg;

    while (PeekMessage (&msg, 0, 0, 0, PM_REMOVE)) {
	TranslateMessage (&msg);
	DispatchMessage (&msg);
    }
}

/* Console Win32 helper routines */
void activate_debugger ();

/* We're not a console-app anymore! */
void setup_brkhandler (void)
{
}

void remove_brkhandler (void)
{
}

int WIN32_RegisterClasses( void )
{
    WNDCLASS wc;

    wc.style = CS_BYTEALIGNCLIENT | CS_BYTEALIGNWINDOW;
    wc.lpfnWndProc = AmigaWindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = 0;
    wc.hIcon = LoadIcon (GetModuleHandle (NULL), MAKEINTRESOURCE( IDI_APPICON ) );
    wc.hCursor = LoadCursor (NULL, IDC_ARROW);
    wc.hbrBackground = GetStockObject (BLACK_BRUSH);
    wc.lpszMenuName = 0;
    wc.lpszClassName = "AmigaPowah";
    if (!RegisterClass (&wc))
	return 0;

    wc.style = CS_BYTEALIGNCLIENT | CS_BYTEALIGNWINDOW;
    wc.lpfnWndProc = MainWindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = 0;
    wc.hIcon = LoadIcon (GetModuleHandle (NULL), MAKEINTRESOURCE( IDI_APPICON ) );
    wc.hCursor = LoadCursor (NULL, IDC_ARROW);
    wc.hbrBackground = GetStockObject (BLACK_BRUSH);
    wc.lpszMenuName = 0;
    wc.lpszClassName = "PCsuxRox";
    if (!RegisterClass (&wc))
	return 0;
    return 1;
}

/* Mathias says: "this truly sucks, I'll include a native gunzip() routine soon" */

static char *uncompress_error[2] = { "Error: You need gzip.exe (32 bit) to use .adz/.roz files!\n", 
                                    "Error: You need xdms.exe (32 bit) to use .dms files!\n" };

int uncompress_hack( int type, const char *src, const char *dst)
{
    int result = 0;
    char fullname[1024];
    char buf[1024];
    char cmd[256];
    char *posn = NULL;
    STARTUPINFO si = {sizeof si};
    PROCESS_INFORMATION pi;

    strcpy( fullname, dst );

    if( type == 1 )
    {
        strcpy( cmd, "gzip.exe -f -d" );
        strcat( fullname, ".gz" );
    }
    else if( type == 2 )
    {
        strcpy( cmd, "xdms.exe u" );
        posn = strrchr( fullname, '.' );
        if( posn )
        {
            *posn = 0;
            strcat( fullname, ".dms" );
        }
    }

    if( CopyFile( src, fullname, FALSE ) ) 
    {
	sprintf (buf, "%s %s +%s", cmd, fullname, dst );
	si.dwFlags = STARTF_USESTDHANDLES;
	if( CreateProcess( NULL, buf, NULL, NULL, TRUE, 0, NULL, NULL, &si, &pi ) ) 
        {
	    WaitForSingleObject( pi.hProcess, INFINITE );
            result = -1;
    	}
        else
        {
            gui_message( uncompress_error[type-1] );
        }
    }

    /* Special handling for broken xdms.exe */
    if( type == 2 )
    {
        DeleteFile( fullname );     /* Delete the uaeXX.dms file */
    }

    return result;
}

#ifdef __GNUC__
#undef WINAPI
#define WINAPI
#endif

static HINSTANCE hRichEdit = NULL, hHtmlHelp = NULL;

int WIN32_CleanupLibraries( void )
{
    if (hRichEdit)
	    FreeLibrary (hRichEdit);
    if (hDDraw)
	    FreeLibrary (hDDraw);
    if( hHtmlHelp )
        FreeLibrary( hHtmlHelp );
    return 1;
}

/* HtmlHelp Initialization - optional component */
int WIN32_InitHtmlHelp( void )
{
    int result = 0;
    if( hHtmlHelp = LoadLibrary( "HHCTRL.OCX" ) )
    {
        pHtmlHelp = ( HWND(WINAPI *)(HWND, LPCSTR, UINT, LPDWORD ) )GetProcAddress( hHtmlHelp, "HtmlHelpA" );
        result = 1;
    }
    return result;
}

/* try to load COMDLG32 and DDRAW, initialize csDraw */
int WIN32_InitLibraries( void )
{
    /* Determine our processor speed and capabilities */
    figure_processor_speed ();

    /* Make sure we do an InitCommonControls() to get some advanced controls */
    InitCommonControls ();

    hRichEdit = LoadLibrary( "RICHED32.DLL" );

    /* DirectDraw is a necessity */
    hDDraw = LoadLibrary ("DDRAW.DLL");
    if (hDDraw == 0) 
    {
        gui_message("You have to install DirectX on your system before you can use UAE.\n"
		            "Refer to the documentation for further details.\n");
	    return 0;
    }
    pDirectDrawCreate = (HRESULT (WINAPI *) (GUID FAR *, LPDIRECTDRAW FAR *, IUnknown FAR *)) GetProcAddress (hDDraw, "DirectDrawCreate");

    return 1;
}

/* console window for debugging messages */
/* Brian: disable for release version if you want (TW) */

#define WRITE_LOG_BUF_SIZE 4096

static int consoleopen;
HANDLE stdinput,stdoutput;

static void openconsole(void)
{
    if(consoleopen) return;
    AllocConsole();
    stdinput=GetStdHandle(STD_INPUT_HANDLE);
    stdoutput=GetStdHandle(STD_OUTPUT_HANDLE);
    SetConsoleMode(stdinput,ENABLE_PROCESSED_INPUT|ENABLE_LINE_INPUT|ENABLE_ECHO_INPUT|ENABLE_PROCESSED_OUTPUT);
}

/* console functions for debugger */

void console_out (const char *format,...)
{
    va_list parms;
    char buffer[WRITE_LOG_BUF_SIZE];
    DWORD temp;

    va_start (parms, format);
    _vsnprintf( buffer, WRITE_LOG_BUF_SIZE-1, format, parms );
    va_end (parms);
    openconsole();
    WriteConsole(stdoutput,buffer,strlen(buffer),&temp,0);
}

int console_get (char *out, int maxlen)
{
    DWORD len,totallen;

    totallen=0;
    while(maxlen>0) 
    {
	    ReadConsole(stdinput,out,1,&len,0);
	    if(*out == 13) break;
	    out++;
	    maxlen--;
	    totallen++;
    }
    *out=0;
    return totallen;
}

void console_flush (void)
{
}

/* GCC/EGCS wants this write_log in order to work from socket-land and to do traces */
#ifdef __GNUC__
void write_log (const char *format,...)
{
    DWORD numwritten;
    char buffer[12];
    va_list parms;
    int count = 0;
    int *blah = NULL;

    if( debugfile )
    {
#if defined HAVE_GETTICKCOUNT && defined TIMESTAMP_LOGS
        {
            sprintf( buffer, "%7d - ", GetTickCount() );
            fprintf( debugfile, buffer );
        }
#endif
        va_start (parms, format);
        count = vfprintf( debugfile, format, parms );
        fflush( debugfile );
        if( count >= WRITE_LOG_BUF_SIZE-1 )
        {
            fprintf( debugfile, "SHIT in write_log()\n" );
            fflush( debugfile );
            *blah = 0; /* Access Violation here! */
            abort();
        }
        va_end (parms);
    }
}
#else /* MSVC likes this one, and so do I */
void write_log( const char *format, ... )
{
    DWORD numwritten;
    char buffer[WRITE_LOG_BUF_SIZE];
    va_list parms;
    int count = 0;
    int *blah = (int *)0xdeadbeef;

    va_start (parms, format);
    count = _vsnprintf( buffer, WRITE_LOG_BUF_SIZE-1, format, parms );
#if defined HAVE_GETTICKCOUNT && defined TIMESTAMP_LOGS
    {
        char buffme[WRITE_LOG_BUF_SIZE];
        sprintf( buffme, "%7d - %s", GetTickCount(), buffer );
        OutputDebugString( buffme );
        if( debugfile )
            fprintf( debugfile, buffme );
    }
#else
    OutputDebugString( buffer );
    if( debugfile )
        fprintf( debugfile, buffer );
#endif
    openconsole();
    WriteConsole(stdoutput,buffer,strlen(buffer),&numwritten,0);
    va_end (parms);
}
#endif

int debuggable (void)
{
    return 0;
}

int needmousehack (void)
{
    return 1;
}

void LED (int a)
{
}

/* report reads/writes to location zero */
void do_enforcer_handling( int read_access, uae_u32 location )
{
    write_log( "An Amiga program just %s location %p!\n", read_access ? "READ" : "WROTE", location );
    FlashWindow( hAmigaWnd, TRUE );
    MessageBeep( MB_OK );
    m68k_dumpstate(NULL,1);
}

void logging_init( void )
{
    char debugfilename[MAX_PATH];
    if( currprefs.win32_logfile )
    {
        sprintf( debugfilename, "%s\\winuaelog.txt", start_path );
        if( !debugfile )
            debugfile = fopen( debugfilename, "wt" );
    }
    write_log (PROGNAME"\n");
    write_log ("\n(c) 1995-1999 Bernd Schmidt   - Core UAE concept and implementation."
	       "\n(c) 1996-1999 Mathias Ortmann - Win32 port and bsdsocket support."
	       "\n(c) 1996-1999 Brian King      - Win32 port, Picasso96 RTG, and GUI."
          "\n(c) 1998-1999 Toni Wilen      - AGA chipset, NTSC/PAL modes."
          "\n(c) 2000      Vasyl Tsvirkunov- Win32 port update.\n"
	       "\nPress F12 to show the Settings Dialog (GUI), Alt-F4 to quit."
           "\nEnd+F1 changes floppy 0, End+F2 changes floppy 1, etc.\n"
	       "\nhttp://www.codepoet.com/UAE/\n\n");
}

void logging_cleanup( void )
{
    if( debugfile )
        fclose( debugfile );
}

static const char *sound_styles[] = { "waveout_looping", "waveout_dblbuff", "dsound_looping", "dsound_dblbuff", 0 };

void target_save_options (FILE *f, struct uae_prefs *p)
{
    fprintf (f, "win32.middle_mouse=%s\n", p->win32_middle_mouse ? "true" : "false");
    fprintf (f, "win32.logfile=%s\n", p->win32_logfile ? "true" : "false");
    fprintf (f, "win32.sound_style=%s\n", sound_styles[ p->win32_sound_style ] );
    fprintf (f, "win32.serial_port=%s\n", p->use_serial ? p->sername : "none" );
    fprintf (f, "win32.parallel_port=%s\n", p->prtname[0] ? p->prtname : "none" );
    fprintf (f, "win32.iconified_nospeed=%s\n", p->win32_iconified_nospeed ? "true" : "false");
    fprintf (f, "win32.iconified_nosound=%s\n", p->win32_iconified_nosound ? "true" : "false");
    fprintf (f, "win32.midiout_device=%d\n", p->win32_midioutdev );
    fprintf (f, "win32.midiin_device=%d\n", p->win32_midiindev );
    fprintf (f, "win32.directx6=%s\n", p->win32_directdraw6 ? "true" : "false" );
}

int target_parse_option (struct uae_prefs *p, char *option, char *value)
{
    int result = (cfgfile_yesno (option, value, "middle_mouse", &p->win32_middle_mouse)
        || cfgfile_yesno  (option, value, "logfile", &p->win32_logfile)
        || cfgfile_yesno  (option, value, "networking", &p->socket_emu)
        || cfgfile_yesno  (option, value, "directx6", &p->win32_directdraw6 )
        || cfgfile_yesno  (option, value, "iconified_nospeed", &p->win32_iconified_nospeed)
        || cfgfile_yesno  (option, value, "iconified_nosound", &p->win32_iconified_nosound)
        || cfgfile_strval (option, value, "sound_style", &p->win32_sound_style, sound_styles, 1)
        || cfgfile_intval (option, value, "sound_tweak", &p->win32_sound_tweak, 1) 
        || cfgfile_intval (option, value, "midi_device", &p->win32_midioutdev, 1)
        || cfgfile_intval (option, value, "midiout_device", &p->win32_midioutdev, 1)
        || cfgfile_intval (option, value, "midiin_device", &p->win32_midiindev, 1)
        || cfgfile_string (option, value, "serial_port", &p->sername[0], 256)
        || cfgfile_string (option, value, "parallel_port", &p->prtname[0], 256));

    if( p->sername[0] == 'n' )
        p->use_serial = 0;
    else
        p->use_serial = 1;

    if( p->prtname[0] == 'n' )
        p->prtname[0] = '\0';

    return result;
}

void machdep_init (void)
{
}

#if defined NO_MAIN_IN_MAIN_C
int PASCAL WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine,
		    int nCmdShow)
{
    hInst = hInstance;

    /*initgfxspecs(); */
#ifdef __CYGWIN32__
    {
	char *token, *tmpbuf;
	int my_argc, i;
	char ** my_argv;

	tmpbuf = my_strdup (lpCmdLine);
	my_argc = 1;
	if (strtok (tmpbuf, "\n ") != NULL) {
	    do {
		my_argc++;
	    } while (strtok (NULL, "\n ") != NULL);
	}
	free (tmpbuf);

	tmpbuf = my_strdup (lpCmdLine);

	my_argv = (char **)xmalloc ((1 + my_argc) * sizeof (char **));
	my_argv[0] = "uae.exe";

	token = strtok (tmpbuf, "\n ");
	i = 1;
	while (token != NULL) {
	    my_argv[i++] = my_strdup (token);
	    token = strtok (NULL, "\n ");
	}
	my_argv[my_argc] = NULL;

	real_main (my_argc, my_argv);
    }
#elif defined __MINGW32__
    real_main (_argc, _argv);
#else
    real_main (__argc, __argv);
#endif
    return FALSE;
}
#endif
